/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */

#include "link.hpp"
#include "pkt.hpp"
#include "router.hpp"

Link::Link(pConnector src, pConnector dst, int d) : Connector(src, dst) {
    delay = d;
    e_bit = param.link_ebit;
}



// Link is a dumb device, so it does not initiate the sending itself, instead
// it will just pass whatever data to the receiver of the link
bool Link::can_send(void) const {
    return source->can_send();
}


bool Link::can_receive(void) const {
    return sink->can_receive();
}


// Link is a passive device, so it will not initiate a send action at all by itself.
// Instead, as soon as it receives a flit, it sends it out immediately. So the following
// function should never be invoked
bool Link::send(void) {
#ifdef DEBUG
    assert(0);
#endif
    return true;
}


// Immediately after the link receive the flit, it sends out the flit instanenously.
// When simulating systems with link error, here is the place to inject/modify links.
// Each power module in the router will take care of all the incoming links power 
// consumption together with the power consumption of the outgoing link to the local
// PE
bool Link::receive(class Flit * a_flit) {
    sink->receive(a_flit);

    pRouter router = 0;

    if (net_clock.get_clock() > param.warmup_period) {
        int dir = -1;

        if (dynamic_cast<Traffic_source *>(source))  // local link does consume zero energy
            return true;

        if (dynamic_cast<In_port *>(sink)) {
            pIn_port p = (pIn_port) sink;
            router = p->get_router();
            dir = (int) p->get_direction();
        }
        /*
          else if (dynamic_cast<Traffic_source *>(source)) {
          assert(dynamic_cast<Out_port *>(source));
          pOut_port p = (pOut_port) source;
          router = p->get_router();
          dir = param.n_of_ports;
          }
        */

        // the link energy consumption of a flit from local outport to local PE is zero
        if (router) {
            param.energy.link_energy += e_bit * a_flit->get_flit_size();
#ifdef USE_ORION
            if (param.orion_power.use_orion_power_model) {
                pPower_module pm = router->get_power_module();
                pm->power_link_traversal(dir, a_flit->raw());
            }
#endif
        }
    }

    return true;
}
