/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */

#include "port.hpp"
#include "pkt.hpp"

Port::~Port() {
    for (vector<pChannel>::iterator iter=channels.begin(); iter<channels.end(); iter++) 
        delete (*iter);
}

void Port::tick(void) {
    for (vector<pChannel>::iterator iter=channels.begin(); iter<channels.end(); iter++) 
        (*iter)->tick();
}

void Port::set_channel_buffer_size(int s) {
    for (vector<pChannel>::iterator iter=channels.begin(); iter<channels.end(); iter++) 
        (*iter)->set_buffer_size(s);
}

// return the sum of the empty buffer slots of all the channels beglong to this port
unsigned int Port::num_of_empty_buffer_slots(void) const {
    unsigned int sum = 0;
    for (vector<pChannel>::const_iterator iter=channels.begin(); iter<channels.end(); iter++)
        sum += (*iter)->num_of_empty_buffer_slots();
    return sum;
}

// return the buffer capacity which is the sum of all input channels' buffer size of this port
unsigned int Port::total_buffer_capacity(void) const {
    unsigned int sum = 0;
    for (vector<pChannel>::const_iterator iter=channels.begin(); iter<channels.end(); iter++)
        sum += (*iter)->get_capacity();
    return sum;
}

In_port::In_port(class Router * r, Direction d) : Port(r, d) {
    pNetwork network = r->get_network();
    size = network->get_num_of_vcs();

    channels.clear();
    Position channel_pos;
    channel_pos.x = d;
    // override buffer size if it's local 
    int buffer_size = param.in_channel_buffer_size;
    //     if (d == local) 
    //         buffer_size = 100;
    for (unsigned int i=0; i<size; i++) {
        channel_pos.y = i;
        pInput_channel a_chan = new Input_channel(r, this, channel_pos, buffer_size);
        channels.push_back(a_chan);
        a_chan->deprive_link();
    }

    // at the beginning, the link is granted to channel0
    channels[0]->grant_link();
}

void In_port::set_routing_scheme(Routing_scheme rs) {
    for (vector<pChannel>::const_iterator iter=channels.begin(); iter<channels.end(); iter++) {
        pInput_channel a_channel = (pInput_channel) *iter;
        pRouting_engine re = a_channel->get_routing_engine();
        re->set_routing_scheme(rs);
    }
}

void In_port::set_routing_table(char * table_string) { //UYO
    for (vector<pChannel>::const_iterator iter=channels.begin(); iter<channels.end(); iter++) {
        pInput_channel a_channel = (pInput_channel) *iter;
        pRouting_engine re = a_channel->get_routing_engine();
        re->set_routing_table(table_string);
    }
}

// whether this port has any data to send to the out port via switch
bool In_port::can_send(void) const {
    for (vector<pChannel>::const_iterator iter=channels.begin(); iter<channels.end(); iter++) 
        if ((*iter)->can_send())
            return true;
    return false;
}

// whether the current link user can receive or not
bool In_port::can_receive(void) const {
    pInput_channel active_channel = (pInput_channel) channels[link_user_channel_id];
    return active_channel->can_receive();
}


bool In_port::send(void) {
    for (vector<pChannel>::const_iterator iter=channels.begin(); iter<channels.end(); iter++) 
        if ((*iter)->can_send()) {
            (*iter)->send();
        }
    return true;
}


bool In_port::receive(pFlit a_flit) {
#ifdef DEBUG
    assert(can_receive());
#endif

    pInput_channel active_channel = (pInput_channel) channels[link_user_channel_id];
    return active_channel->receive(a_flit);
}


Out_port::Out_port(class Router * r, Direction d) : Port(r, d) {
    pNetwork network = r->get_network();
    size = network->get_num_of_vcs();

    channels.clear();
    Position channel_pos;
    channel_pos.x = d;
    int buffer_size = param.out_channel_buffer_size;
    if (d == local && param.out_channel_buffer_size < 2)
        buffer_size = 2;   // out channel should be at least 2 in order to mimic infinite local out buffer
    for (unsigned int i=0; i<size; i++) {
        channel_pos.y = i;
        pOutput_channel a_chan = new Output_channel(r, this, channel_pos, buffer_size);
        channels.push_back(a_chan);
        a_chan->deprive_link();
    }

    // channel 0 has the access to the link after initialization
    channels[0]->grant_link();
}


// whether the current link user can send data or not
bool Out_port::can_send(void) const {
    pOutput_channel active_channel = (pOutput_channel) channels[link_user_channel_id];
    return active_channel->can_send();
}

// whether this port's channel can receive any data via switch
bool Out_port::can_receive(void) const {
    for (vector<pChannel>::const_iterator iter=channels.begin(); iter<channels.end(); iter++) 
        if ((*iter)->can_receive())
            return true;
    return false;
}


bool Out_port::send(void) {
    pOutput_channel active_channel = (pOutput_channel) channels[link_user_channel_id];
    return active_channel->send();
}


// do nothing as the receiving at the granularity of each channel for out port
bool Out_port::receive(pFlit a_flit) {
    return true;
}
